require "fiddle"
require "rbconfig"
require_relative "bearlib"

module Bearlib
  module Raw
    # Raw API entries
    Alloc = Fiddle::Function.new(Libmap["map_alloc"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_VOIDP], Fiddle::TYPE_INT)
    Free = Fiddle::Function.new(Libmap["map_free"], [Fiddle::TYPE_INT], Fiddle::TYPE_VOID)
    Copy = Fiddle::Function.new(Libmap["map_copy"], [Fiddle::TYPE_INT], Fiddle::TYPE_INT)
    Assign = Fiddle::Function.new(Libmap["map_assign"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT], Fiddle::TYPE_INT)
    Clear = Fiddle::Function.new(Libmap["map_clear"], [Fiddle::TYPE_INT], Fiddle::TYPE_VOID)
    ClearLayer = Fiddle::Function.new(Libmap["map_clear_layer"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT], Fiddle::TYPE_VOID)
    Set = Fiddle::Function.new(Libmap["map_set"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT], Fiddle::TYPE_VOID)
    SetF = Fiddle::Function.new(Libmap["map_setf"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_FLOAT], Fiddle::TYPE_VOID)
    Get = Fiddle::Function.new(Libmap["map_get"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT], Fiddle::TYPE_INT)
    GetF = Fiddle::Function.new(Libmap["map_getf"], [Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT, Fiddle::TYPE_INT], Fiddle::TYPE_FLOAT)
    Width = Fiddle::Function.new(Libmap["map_width"], [Fiddle::TYPE_INT], Fiddle::TYPE_INT)
    Height = Fiddle::Function.new(Libmap["map_height"], [Fiddle::TYPE_INT], Fiddle::TYPE_INT)
  end

  # Wrappers

  LAYER_TYPES = [:flag, :integer, :float]

  class Layer
    attr_reader :owner, :id

    def initialize(owner, id)
      @owner = owner
      @id = id
    end

    def clear
      Raw::ClearLayer.call(@owner.handle, id)
    end

    def [](x, y)
      Raw::Get.call(@owner.handle, x, y, id)
    end

    def []=(x, y, v)
      Raw::Set.call(@owner.handle, x, y, id, v)
    end
  end

  class FloatLayer < Layer
    def [](x, y)
      Raw::Getf.call(@owner.handle, x, y, id)
    end

    def []=(x, y, v)
      Raw::Setf.call(@owner.handle, x, y, id, v)
    end
  end

  class BoolLayer < Layer
    def [](x, y)
      Raw::Get.call(@owner.handle, x, y, id) != 0
    end

    def []=(x, y, v)
      Raw::Set.call(@owner.handle, x, y, id, v ? 1 : 0)
    end
  end

  LAYER_TYPES_TO_CLASSES = {:flag => BoolLayer, :integer => Layer, :float => FloatLayer}

  class Map
    attr_reader :handle, :width, :height, :layers

    def initialize(*args)
      if args.count > 0
        width, height, layer_ids = args
        if layer_ids.is_a? String
          layer_ids = layer_ids.split(",").map(&:strip).map(&:to_sym)
        end
        if layer_ids.is_a? Hash
          @layers = {}
          hash = layer_ids
          layer_ids = []
          hash.each_with_index do |(k, v), i|
            @layers[k] = LAYER_TYPES_TO_CLASSES[v].new(self, i)
            layer_ids << v
          end
        else
          @layers = layer_ids.map_with_index { |s, i| LAYER_TYPES_TO_CLASSES[s].new(self, i) }
        end

        @width = width
        @height = height
        @handle = Raw::Alloc.call width, height, Raw::Ptr[layer_ids.map(&:to_s).join(",")]
      end
    end

    def free
      Raw::Free.call @handle
    end

    def clone_from(amap)
      @width = amap.width
      @height = amap.height
      @layers = amap.layers
      @handle = Raw::Copy.call(amap.handle)
      self
    end

    def clone
      Raw::Map.new.clone_from(amap)
    end

    def assign(from_map)
      Raw::Assign.call(@handle, from_map.handle)
    end

    def clear
      Raw::Clear.call(@handle)
    end
  end
end
